import pytest

from raesl.doc.cli import doc


def test_doc_cli(tmpdir, datadir):
    datadir = datadir.absolute()
    with tmpdir.as_cwd():
        esl_paths = str(datadir / "specs/pump_example.esl")
        with pytest.raises(SystemExit) as e:
            doc([esl_paths, "--dry"])
        assert e.value.code == 0



"""Tests for conversion to Excel module."""

from pathlib import Path

import openpyxl

import raesl.excel


def test_excel(datadir: Path, update, tmpdir):
    """Integral test for Excel module."""
    spec = datadir / "specs" / "pump_example.esl"

    ref = datadir / "excel" / "pump.xlsx"
    output = ref if update else tmpdir / "pump.xlsx"
    raesl.excel.convert(spec, output=output)

    if update:
        return

    wb = openpyxl.load_workbook(output)
    wb_ref = openpyxl.load_workbook(ref)

    # This is in no means comprehensive, but serves as a global output test.
    assert wb.sheetnames == wb_ref.sheetnames
    for ws, ws_ref in zip(wb.worksheets, wb_ref.worksheets):
        for col, col_ref in zip(ws.iter_cols(), ws_ref.iter_cols()):
            for cell, cell_ref in zip(col, col_ref):
                assert cell.value == cell_ref.value
                assert cell.style == cell_ref.style


def test_component_overview(gendir: Path, update, tmpdir):
    """Integral test for component overview generation."""

    from raesl import datasets
    from raesl.excel import component_overview

    graph = datasets.get("rally-car")

    for comp in graph.get_nodes_by_kind("component"):
        ref = gendir / f"overview_{comp.name}.xlsx"
        output = ref if update else tmpdir / f"overview_{comp.name}.xlsx"

        component_overview(graph, comp.name, output=output)



"""Test compiler output for specifications in data/specs"""

import difflib
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Optional

from ragraph.io.json import from_json, to_json

import raesl.plot
from raesl.compile.ast.specification import dump
from raesl.compile.cli import run

if TYPE_CHECKING:
    from raesl.compile.ast.specification import Specification
    from raesl.compile.diagnostics import DiagnosticStore


def compare_texts(exp_path: Path, real_path: Path, update) -> Optional[Iterable[str]]:
    """Compare expected output with real output, and display differences.

    Returns:
        Diff lines or None.
    """
    real_text = real_path.read_text() if real_path.exists() else ""
    if update:
        if real_text == "":
            if exp_path.exists():
                exp_path.unlink()
        else:
            exp_path.write_text(real_text)
        return None
    exp_text = exp_path.read_text() if exp_path.exists() else ""

    if exp_text == real_text:
        return None

    exp_lines = exp_text.split("\n")
    real_lines = real_text.split("\n")
    diff = difflib.unified_diff(
        exp_lines,
        real_lines,
        fromfile=str(exp_path),
        tofile=str(real_path),
        lineterm="",
    )
    return diff


def check_spec(
    specification: Optional["Specification"],
    spec_path: Path,
    tmpdir: Path,
    update: bool,
):
    """Check specification output."""
    exp_path = spec_path.with_suffix(".exp_output")
    real_path = tmpdir / (spec_path.stem + ".output")
    if specification is not None:
        with real_path.open("a") as f:
            dump(specification, f)
    else:
        real_path.write_text("")
    diff = compare_texts(exp_path, real_path, update)
    assert diff is None, "\n".join(diff)


def check_diag(diag_store: "DiagnosticStore", spec_path: Path, tmpdir: Path, update: bool):
    """Check diagnostic output."""
    exp_path = spec_path.with_suffix(".exp_error")
    real_path = tmpdir / (spec_path.stem + ".error")
    with real_path.open("a") as f:
        diag_store.dump(test=True, stream=f)
    diff = compare_texts(exp_path, real_path, update)
    assert diff is None, "\n".join(diff)


def test_doccomment_handling(datadir: Path, update: bool):
    """Test for multi-file doc comment handling"""

    fpath = datadir / "doccomments"

    diag_store, parsed_spec, g = run(fpath)
    assert g is not None

    assert g["world.x"].annotations.esl_info["comments"] == [
        "Oh yeah it works!",
        "Even multi-line!",
    ]
    assert g["world.blib"].annotations.esl_info["comments"] == [
        "Hello",
    ]
    assert g["world.blib.A"].annotations.esl_info["comments"] == [
        "Apple",
    ]
    assert g["world.blib.B"].annotations.esl_info["comments"] == [
        "Pie",
    ]
    assert g["world.blib.g1"].annotations.esl_info["comments"] == [
        "Strawberry",
        "Bonjour",
    ]
    assert g["world.blib.g1"].annotations.esl_info["comments"] == [
        "Strawberry",
        "Bonjour",
    ]
    assert g["world.blib.g2"].annotations.esl_info["comments"] == [
        "Banana's",
        "Au revoir",
    ]
    assert g["world.blub"].annotations.esl_info["comments"] == [
        "Goodby",
    ]
    assert g["world.blub.A"].annotations.esl_info["comments"] == [
        "Apple",
    ]
    assert g["world.blub.B"].annotations.esl_info["comments"] == [
        "Pie",
    ]
    assert g["world.blub.g1"].annotations.esl_info["comments"] == [
        "Strawberry",
        "Bonjour",
    ]
    assert g["world.blub.g1"].annotations.esl_info["comments"] == [
        "Strawberry",
        "Bonjour",
    ]
    assert g["world.blub.g2"].annotations.esl_info["comments"] == [
        "Banana's",
        "Au revoir",
    ]
    assert g["world.g2"].annotations.esl_info["comments"] == ["Peace", "Au revoir"]
    assert g["world.g4"].annotations.esl_info["comments"] == ["Pear", "Bonjour"]


def test_graph_building(datadir: Path, update: bool, inspect: bool):
    fpaths = [
        datadir / "specs" / "pump_example.esl",
        datadir / "specs" / "behavior-requirements_path_dependencies.esl",
        datadir / "specs" / "design_dependency_migration.esl",
    ]

    for fpath in fpaths:
        diag_store, parsed_spec, g = run(fpath)
        assert g is not None

        if diag_store.diagnostics:
            diag_store.dump()

        ref_file = fpath.with_suffix(".json")

        if update:
            to_json(g, ref_file)

        g_ref = from_json(ref_file)

        for n in g.nodes:
            assert g[n.name].annotations.esl_info == g_ref[n.name].annotations.esl_info

        for ni in g.nodes:
            for nj in g.nodes:
                for e, e_ref in zip(g[ni.name, nj.name], g_ref[ni.name, nj.name]):
                    if hasattr(e, "esl_info"):
                        for key, val in e.annotations.esl_info["reason"]:
                            assert val.sort == e_ref.annotations.esl_inf["reason"]["key"].sort()

        if inspect:
            style = raesl.plot.Style(ragraph=dict(piemap={"display": "labels", "mode": "relative"}))
            fig = raesl.plot.mdm(
                graph=g,
                node_kinds=[
                    "component",
                    "function_spec",
                    "behavior_spec",
                    "variable",
                    "design_spec",
                    "relation_spec",
                    "need",
                ],
                edge_kinds=[
                    "functional_dependency",
                    "mapping_dependency",
                    "logical_dependency",
                    "coordination_dependency",
                    "design_dependency",
                ],
                depth=10,
                style=style,
            )
            fig.show()


def test_specs(spec: Path, datadir: Path, tmpdir: Path, update: bool):
    diag_store, compiled_spec, graph = run(spec)
    check_spec(compiled_spec, spec, Path(tmpdir), update)
    check_diag(diag_store, spec, Path(tmpdir), update)


def test_bad_specs(bad_example: Path, datadir: Path, tmpdir: Path, update: bool):
    diag_store, compiled_spec, graph = run(bad_example)
    check_spec(compiled_spec, bad_example, Path(tmpdir), update)
    check_diag(diag_store, bad_example, Path(tmpdir), update)


def test_good_specs(good_example: Path, datadir: Path, tmpdir: Path, update: bool):
    diag_store, compiled_spec, graph = run(good_example)
    check_spec(compiled_spec, good_example, Path(tmpdir), update)
    check_diag(diag_store, good_example, Path(tmpdir), update)



from pathlib import Path

from raesl.doc.doc import Doc
from raesl.utils import get_esl_paths

LOCAL_DEBUG = False


def test_esls_en(tmpdir, datadir, gendir, inspect):
    datadir = datadir.absolute()
    gendir = gendir.absolute()

    with tmpdir.as_cwd():
        d = Doc(
            datadir / "doc" / "esl" / "pump.esl",
            prologue=datadir / "doc" / "md" / "prologue.md",
            epilogue=datadir / "doc" / "md" / "epilogue.md",
            rich="tex",
            language="en",
            title="Pump",
        )
        if inspect:
            out_file = Path(gendir) / "pump_en.pdf"
        else:
            out_file = "pump_en.pdf"
        d.save(out_file)


def test_esls_nl(tmpdir, datadir, gendir, inspect):
    datadir = datadir.absolute()
    gendir = gendir.absolute()

    for esl_paths, name in zip(
        [
            [datadir / "doc" / "esl" / "basisspecificatie.esl"],
            get_esl_paths(datadir / "doc" / "esl" / "noodstopketen"),
        ],
        ["Basisspecificatie", "Noodstopketen"],
    ):
        with tmpdir.as_cwd():
            d = Doc(
                *esl_paths,
                prologue=datadir / "doc" / "md" / "prologue.md",
                epilogue=datadir / "doc" / "md" / "epilogue.md",
                rich="tex",
                language="nl",
                title=name,
            )
            if inspect:
                out_file = Path(gendir) / f"{name}.pdf"
            else:
                out_file = f"{name}.pdf"
            d.save(out_file)


def test_pump_md(tmpdir, datadir, gendir, inspect):
    datadir = datadir.absolute()
    gendir = gendir.absolute()

    with tmpdir.as_cwd():
        d = Doc(
            datadir / "doc" / "esl" / "pump.esl",
            prologue=datadir / "doc" / "md" / "prologue.md",
            epilogue=datadir / "doc" / "md" / "epilogue.md",
            rich="md",
            language="en",
            title="Pump",
        )
        if inspect:
            out_file = Path(gendir) / "pump_en.pdf"
        else:
            out_file = "pump_en.pdf"
        d.save(out_file)


def test_pump_set_mdm(tmpdir, datadir, gendir, inspect):
    datadir = datadir.absolute()
    gendir = gendir.absolute()

    with tmpdir.as_cwd():
        d = Doc(
            datadir / "doc" / "esl" / "pump.esl",
            prologue=datadir / "doc" / "md" / "prologue.md",
            epilogue=datadir / "doc" / "md" / "epilogue.md",
            rich="tex",
            rich_opts={
                "node_kinds": ["component"],
                "edge_kinds": ["functional_dependency"],
                "pie_mode": "relative",
            },
            language="en",
            title="Pump",
        )
        if inspect:
            out_file = Path(gendir) / "pump_set_mdm.pdf"
        else:
            out_file = "pump_set_mdm.pdf"
        d.save(out_file)


def test_bundle_handling(tmpdir, datadir, gendir, inspect):
    datadir = datadir.absolute()
    gendir = gendir.absolute()

    with tmpdir.as_cwd():
        d = Doc(
            datadir / "doc/esl/bundle_handling.esl",
            rich="tex",
            language="en",
            title="Bundle handling",
        )
        if inspect:
            out_file = Path(gendir) / "bundle_handling.pdf"
        else:
            out_file = "bundle_handling.pdf"
        d.save(out_file)


def test_bundle_handling_nl(tmpdir, datadir, gendir, inspect):
    datadir = datadir.absolute()
    gendir = gendir.absolute()

    with tmpdir.as_cwd():
        d = Doc(
            datadir / "doc/esl/bundle_handling.esl",
            rich="tex",
            language="nl",
            title="Bundle handling",
        )
        if inspect:
            out_file = Path(gendir) / "bundle_handling_nl.pdf"
        else:
            out_file = "bundle_handling_nl.pdf"
        d.save(out_file)



from ragraph import colors
from ragraph.analysis.heuristics import markov_gamma
from ragraph.graph import Graph
from ragraph.io.esl import from_esl

import raesl.plot


def test_mdm_0(pump_example_graph: Graph, check_plotly):
    """Testing the creating of a multi-domain-matrix plot."""
    g = pump_example_graph

    style = raesl.plot.Style(ragraph=dict(piemap={"display": "labels", "mode": "relative"}))

    for depth in [1, 2]:
        mdm = raesl.plot.mdm(
            g,
            node_kinds=["component", "function_spec", "variable"],
            edge_kinds=["functional_dependency", "mapping_dependency"],
            depth=depth,
            style=style,
        )

        fname = f"pump_mdm_fig_level_{depth}.json"
        check_plotly(mdm, fname)


def test_mdm_1(datadir, check_plotly):
    g = from_esl(datadir / "doc" / "esl" / "noodstopketen")

    col_dict = {
        label: color
        for label, color in zip(g.edge_labels, colors.get_categorical(len(g.edge_labels)))
    }

    # Plot clustered matrix
    markov_gamma(
        g,
        alpha=2,
        beta=4.0,
        mu=3.0,
        gamma=2.0,
        local_buses=True,
        leafs=[n for n in g.leafs if n.kind == "component"],
    )

    markov_gamma(
        g,
        alpha=2,
        beta=3.5,
        mu=3.5,
        gamma=10.0,
        local_buses=True,
        leafs=[n for n in g.leafs if n.kind == "function_spec"],
    )

    style = raesl.plot.Style(
        ragraph=dict(
            piemap={"display": "labels", "mode": "relative"},
            palettes={"fields": col_dict},
            show_legend=True,
        )
    )

    depth = g.max_depth
    mdm = raesl.plot.mdm(
        g,
        node_kinds=["component", "function_spec"],
        edge_kinds=["functional_dependency", "mapping_dependency"],
        depth=depth,
        style=style,
    )

    fname = "noodstopketen_mdm_clustered.json"
    check_plotly(mdm, fname)
